/**
 * The plugin that enhances {@link core!ArrayShape ArrayShape} with additional methods.
 *
 * ```ts
 * import { ArrayShape } from 'doubter/core';
 * import enableArrayEssentials from 'doubter/plugin/array-essentials';
 *
 * enableArrayEssentials(ArrayShape);
 * ```
 *
 * @module plugin/array-essentials
 */
import { ArrayShape } from '../shape/ArrayShape';
import { AnyShape } from '../shape/Shape';
import { Any, IssueOptions, Message } from '../types';
declare module '../core' {
    interface Messages {
        /**
         * @default "Must include a value"
         */
        'array.includes': Message | Any;
        /**
         * @default "Must have the maximum length of %s"
         */
        'array.max': Message | Any;
        /**
         * @default "Must have the minimum length of %s"
         */
        'array.min': Message | Any;
    }
    interface ArrayShape<HeadShapes extends readonly AnyShape[], RestShape extends AnyShape | null> {
        /**
         * Constrains the array length.
         *
         * @param length The minimum array length.
         * @param options The issue options or the issue message.
         * @returns The clone of the shape.
         * @group Plugin Methods
         * @plugin {@link plugin/array-essentials! plugin/array-essentials}
         */
        length(length: number, options?: IssueOptions | Message): this;
        /**
         * Constrains the minimum array length.
         *
         * @param length The minimum array length.
         * @param options The issue options or the issue message.
         * @returns The clone of the shape.
         * @group Plugin Methods
         * @plugin {@link plugin/array-essentials! plugin/array-essentials}
         */
        min(length: number, options?: IssueOptions | Message): this;
        /**
         * Constrains the maximum array length.
         *
         * @param length The maximum array length.
         * @param options The issue options or the issue message.
         * @returns The clone of the shape.
         * @group Plugin Methods
         * @plugin {@link plugin/array-essentials! plugin/array-essentials}
         */
        max(length: number, options?: IssueOptions | Message): this;
        /**
         * Constrains the array length to be at least one element.
         *
         * @param options The issue options or the issue message.
         * @returns The clone of the shape.
         * @group Plugin Methods
         * @plugin {@link plugin/array-essentials! plugin/array-essentials}
         */
        nonEmpty(options?: IssueOptions | Message): this;
        /**
         * Requires an array to contain at least one element that conforms the given shape.
         *
         * @param value The shape of the required element or its literal value.
         * @param options The issue options or the issue message.
         * @returns The clone of the shape.
         * @group Plugin Methods
         * @plugin {@link plugin/array-essentials! plugin/array-essentials}
         */
        includes(value: AnyShape | any, options?: IssueOptions | Message): this;
    }
}
/**
 * Enhances {@link core!ArrayShape ArrayShape} with additional methods.
 */
export default function enableArrayEssentials(ctor: typeof ArrayShape<any, any>): void;
//# sourceMappingURL=array-essentials.d.ts.map