/**
 * The plugin that enhances {@link core!BigIntShape BigIntShape} with additional methods.
 *
 * ```ts
 * import { BigIntShape } from 'doubter/core';
 * import enableBigIntEssentials from 'doubter/plugin/bigint-essentials';
 *
 * enableBigIntEssentials(BigIntShape);
 * ```
 *
 * @module plugin/bigint-essentials
 */
import { BigIntShape } from '../shape/BigIntShape';
import { Any, IssueOptions, Message } from '../types';
declare module '../core' {
    interface Messages {
        /**
         * @default "Must be greater than or equal to %s"
         */
        'bigint.min': Message | Any;
        /**
         * @default "Must be less than or equal to %s"
         */
        'bigint.max': Message | Any;
    }
    interface BigIntShape {
        /**
         * Constrains the bigint to be greater than zero.
         *
         * @param options The issue options or the issue message.
         * @returns The clone of the shape.
         * @group Plugin Methods
         * @plugin {@link plugin/bigint-essentials! plugin/bigint-essentials}
         */
        positive(options?: IssueOptions | Message): this;
        /**
         * Constrains the bigint to be less than zero.
         *
         * @param options The issue options or the issue message.
         * @returns The clone of the shape.
         * @group Plugin Methods
         * @plugin {@link plugin/bigint-essentials! plugin/bigint-essentials}
         */
        negative(options?: IssueOptions | Message): this;
        /**
         * Constrains the bigint to be less or equal to zero.
         *
         * @param options The issue options or the issue message.
         * @returns The clone of the shape.
         * @group Plugin Methods
         * @plugin {@link plugin/bigint-essentials! plugin/bigint-essentials}
         */
        nonPositive(options?: IssueOptions | Message): this;
        /**
         * Constrains the bigint to be greater or equal to zero.
         *
         * @param options The issue options or the issue message.
         * @returns The clone of the shape.
         * @group Plugin Methods
         * @plugin {@link plugin/bigint-essentials! plugin/bigint-essentials}
         */
        nonNegative(options?: IssueOptions | Message): this;
        /**
         * Constrains the bigint to be greater than or equal to the value.
         *
         * @param value The inclusive minimum value.
         * @param options The issue options or the issue message.
         * @returns The clone of the shape.
         * @group Plugin Methods
         * @plugin {@link plugin/bigint-essentials! plugin/bigint-essentials}
         */
        min(value: bigint | number | string, options?: IssueOptions | Message): this;
        /**
         * Constrains the number to be less than or equal to the value.
         *
         * @param value The inclusive maximum value.
         * @param options The issue options or the issue message.
         * @returns The clone of the shape.
         * @group Plugin Methods
         * @plugin {@link plugin/bigint-essentials! plugin/bigint-essentials}
         */
        max(value: bigint | number | string, options?: IssueOptions | Message): this;
    }
}
/**
 * Enhances {@link core!BigIntShape BigIntShape} with additional methods.
 */
export default function enableBigIntEssentials(ctor: typeof BigIntShape): void;
//# sourceMappingURL=bigint-essentials.d.ts.map