import { INPUT, OUTPUT } from '../internal/shapes';
import { ApplyOptions, Issue, IssueOptions, Message, Result } from '../types';
import { CoercibleShape } from './CoercibleShape';
import { AnyShape, DeepPartialProtocol, OptionalDeepPartialShape, Shape } from './Shape';
type InferArray<HeadShapes extends readonly AnyShape[], RestShape extends AnyShape | null, Leg extends INPUT | OUTPUT> = [
    ...{
        [K in keyof HeadShapes]: HeadShapes[K][Leg];
    },
    ...(RestShape extends null | undefined ? [] : RestShape extends Shape ? RestShape[Leg][] : [])
];
type DeepPartialArrayShape<HeadShapes extends readonly AnyShape[], RestShape extends AnyShape | null> = ArrayShape<{
    [K in keyof HeadShapes]: OptionalDeepPartialShape<HeadShapes[K]>;
}, RestShape extends null | undefined ? null : RestShape extends Shape ? OptionalDeepPartialShape<RestShape> : RestShape>;
/**
 * The shape of an array or a tuple value.
 *
 * | Shape | Input Type | Output Type |
 * | :-- | :-- | :-- |
 * | `ArrayShape<[], null>` | `[]` | `[]` |
 * | `ArrayShape<[], Shape<A, B>>` | `A[]` | `B[]` |
 * | `ArrayShape<​[Shape<A, B>, Shape<X, Y>], null>` | `[A, X]` | `[B, Y]` |
 * | `ArrayShape<​[Shape<A, B>], Shape<X, Y>>` | `[A, ...X[]]` | `[B, ...Y[]]` |
 *
 * @template HeadShapes The array of positioned element shapes.
 * @template RestShape The shape of rest elements, or `null` if there are no rest elements.
 * @group Shapes
 */
export declare class ArrayShape<HeadShapes extends readonly AnyShape[], RestShape extends AnyShape | null> extends CoercibleShape<InferArray<HeadShapes, RestShape, INPUT>, InferArray<HeadShapes, RestShape, OUTPUT>, unknown[]> implements DeepPartialProtocol<DeepPartialArrayShape<HeadShapes, RestShape>> {
    /**
     * The array of positioned element shapes.
     */
    readonly headShapes: HeadShapes;
    /**
     * The shape of rest elements or `null` if there are no rest elements.
     */
    readonly restShape: RestShape;
    /**
     * The type constraint options or the type issue message.
     */
    protected _options: IssueOptions | Message | undefined;
    /**
     * Returns issues associated with an invalid input value type.
     */
    protected _typeIssueFactory: (input: unknown, options: ApplyOptions) => Issue;
    /**
     * Creates a new {@link ArrayShape} instance.
     *
     * @param headShapes The array of positioned element shapes.
     * @param restShape The shape of rest elements or `null` if there are no rest elements.
     * @param options The issue options or the issue message.
     * @template HeadShapes The array of positioned element shapes.
     * @template RestShape The shape of rest elements, or `null` if there are no rest elements.
     */
    constructor(
    /**
     * The array of positioned element shapes.
     */
    headShapes: HeadShapes, 
    /**
     * The shape of rest elements or `null` if there are no rest elements.
     */
    restShape: RestShape, options?: IssueOptions | Message);
    at(key: unknown): AnyShape | null;
    /**
     * Returns an array shape that has rest elements constrained by the given shape.
     *
     * **Note:** This method returns a shape without any operations.
     *
     * @param restShape The shape of rest elements, or `null` if there are no rest elements.
     * @returns The new array shape.
     * @template S The shape of rest elements.
     */
    rest<S extends AnyShape | null>(restShape: S): ArrayShape<HeadShapes, S>;
    deepPartial(): DeepPartialArrayShape<HeadShapes, RestShape>;
    protected _isAsync(): boolean;
    protected _getInputs(): readonly unknown[];
    /**
     * Coerces a value to an array.
     *
     * @param input The value to coerce.
     * @returns An array, or {@link NEVER} if coercion isn't possible.
     */
    protected _coerce(input: unknown): unknown[];
    protected _apply(input: any, options: ApplyOptions, nonce: number): Result<InferArray<HeadShapes, RestShape, OUTPUT>>;
    protected _applyAsync(input: any, options: ApplyOptions, nonce: number): Promise<Result<InferArray<HeadShapes, RestShape, OUTPUT>>>;
}
export {};
//# sourceMappingURL=ArrayShape.d.ts.map