import { Awaitable, INPUT, OUTPUT, Promisify } from '../internal/shapes';
import { ApplyOptions, IssueOptions, Message, ParseOptions, Result } from '../types';
import { AnyShape, Input, Output, Shape } from './Shape';
type InferOrDefault<Shape extends AnyShape | null, Leg extends INPUT | OUTPUT, DefaultValue = any> = Shape extends null | undefined ? DefaultValue : Shape extends AnyShape ? Shape[Leg] : DefaultValue;
type ThisType<F> = F extends (this: infer T, ...args: any[]) => any ? T : any;
/**
 * The shape of a function.
 *
 * @template ArgsShape The shape of the array of arguments.
 * @template ReturnShape The return value shape, or `null` if unconstrained.
 * @template ThisShape The shape of `this` argument, or `null` if unconstrained.
 * @group Shapes
 */
export declare class FunctionShape<ArgsShape extends Shape<readonly any[], readonly any[]>, ReturnShape extends AnyShape | null, ThisShape extends AnyShape | null> extends Shape<(this: InferOrDefault<ThisShape, OUTPUT>, ...args: Output<ArgsShape>) => InferOrDefault<ReturnShape, INPUT>, (this: InferOrDefault<ThisShape, INPUT>, ...args: Input<ArgsShape>) => InferOrDefault<ReturnShape, OUTPUT>> {
    /**
     * The shape of the array of arguments.
     */
    readonly argsShape: ArgsShape;
    /**
     * The return value shape, or `null` if unconstrained.
     */
    readonly returnShape: ReturnShape;
    /**
     * The shape of `this` value, or `null` if unconstrained.
     */
    readonly thisShape: ThisShape;
    /**
     * `true` if input functions are wrapped during parsing to ensure runtime signature type-safety, or `false` otherwise.
     */
    isStrict: boolean;
    /**
     * Returns issues associated with an invalid input value type.
     */
    protected _typeIssueFactory: (input: unknown, options: ApplyOptions) => import("../types").Issue;
    /**
     * Parsing options that are used by a wrapper.
     */
    protected _parseOptions: ParseOptions | undefined;
    /**
     * Creates a new {@link FunctionShape} instance.
     *
     * @param argsShape The shape of the array of arguments.
     * @param returnShape The return value shape, or `null` if unconstrained.
     * @param thisShape The shape of `this` argument, or `null` if unconstrained.
     * @param options The issue options or the issue message.
     * @template ArgsShape The shape of the array of arguments.
     * @template ReturnShape The return value shape, or `null` if unconstrained.
     * @template ThisShape The shape of `this` argument, or `null` if unconstrained.
     */
    constructor(
    /**
     * The shape of the array of arguments.
     */
    argsShape: ArgsShape, 
    /**
     * The return value shape, or `null` if unconstrained.
     */
    returnShape: ReturnShape, 
    /**
     * The shape of `this` value, or `null` if unconstrained.
     */
    thisShape: ThisShape, options?: IssueOptions | Message);
    /**
     * `true` if some shapes that describe the function signature are {@link Shape.isAsync async}, or `false` otherwise.
     */
    get isAsyncFunction(): boolean;
    /**
     * Constrains the function return value with the given shape.
     *
     * @param shape The return value shape, or `null` if unconstrained.
     * @returns The new function shape.
     * @template S The return value shape.
     */
    return<S extends AnyShape | null>(shape: S): FunctionShape<ArgsShape, S, ThisShape>;
    /**
     * Constrains the type of `this` inside the function.
     *
     * @param shape The shape of `this` argument, or `null` if unconstrained.
     * @returns The new function shape.
     * @template S The shape of `this` argument.
     */
    this<S extends AnyShape | null>(shape: S): FunctionShape<ArgsShape, ReturnShape, S>;
    /**
     * Enables input function wrapping during parsing to ensure runtime signature type-safety. Wrapper ensures that input
     * function receives arguments and `this` values that conform {@link FunctionShape.argsShape} and
     * {@link FunctionShape.thisShape} respectively, and returns the value that conforms
     * {@link FunctionShape.returnShape}.
     *
     * @param options Options that are used by the wrapper. If omitted then default options are applied: no early-return,
     * no type coercion.
     * @returns The new function shape.
     */
    strict(options?: ParseOptions): this;
    /**
     * Creates a function that ensures that `fn` receives arguments and `this` value that conform the
     * {@linkcode FunctionShape.argsShape} and the {@linkcode FunctionShape.thisShape} respectively, and _synchronously_
     * returns the value that conforms the {@linkcode FunctionShape.returnShape}.
     *
     * @param fn The underlying function.
     * @param options Parsing options. By default, options provided to {@linkcode FunctionShape.strict} are used.
     * @returns The wrapper function.
     * @template F The function to which signature must be ensured.
     */
    ensure<F extends (this: InferOrDefault<ThisShape, OUTPUT>, ...args: Output<ArgsShape>) => InferOrDefault<ReturnShape, INPUT>>(fn: F, options?: ParseOptions): (this: InferOrDefault<ThisShape, INPUT, ThisType<F>>, ...args: Input<ArgsShape>) => InferOrDefault<ReturnShape, OUTPUT, ReturnType<F>>;
    /**
     * Creates a function that ensures that `fn` receives arguments and `this` value that conform the
     * {@linkcode FunctionShape.argsShape} and the {@linkcode FunctionShape.thisShape} respectively, and _asynchronously_
     * returns the value that conforms the {@linkcode FunctionShape.returnShape}.
     *
     * Use this method if {@link FunctionShape.isAsyncFunction some shapes that describe the function signature} are
     * {@link Shape.isAsync async}.
     *
     * @param fn The underlying function.
     * @param options Parsing options. By default, options provided to {@linkcode FunctionShape.strict} are used.
     * @returns The wrapper function.
     * @template F The function to which signature must be ensured.
     */
    ensureAsync<F extends (this: InferOrDefault<ThisShape, OUTPUT>, ...args: Output<ArgsShape>) => Awaitable<InferOrDefault<ReturnShape, INPUT>>>(fn: F, options?: ParseOptions): (this: InferOrDefault<ThisShape, INPUT, ThisType<F>>, ...args: Input<ArgsShape>) => Promisify<InferOrDefault<ReturnShape, OUTPUT, ReturnType<F>>>;
    protected _getInputs(): readonly unknown[];
    protected _apply(input: any, options: ApplyOptions, nonce: number): Result<(this: InferOrDefault<ThisShape, INPUT>, ...args: Input<ArgsShape>) => InferOrDefault<ReturnShape, OUTPUT>>;
    protected _applyAsync(input: any, options: ApplyOptions, nonce: number): Promise<Result<(this: InferOrDefault<ThisShape, INPUT>, ...args: Input<ArgsShape>) => InferOrDefault<ReturnShape, OUTPUT>>>;
}
export {};
//# sourceMappingURL=FunctionShape.d.ts.map