import { Any, ApplyOptions, Result } from '../types';
import { AnyShape, DeepPartialProtocol, DeepPartialShape, Input, Output, Shape } from './Shape';
/**
 * Lazily loads a shape using the provider callback.
 *
 * @template ProvidedShape The lazy-loaded shape.
 * @template Pointer The value returned when a cyclic reference is detected.
 * @group Shapes
 */
export declare class LazyShape<ProvidedShape extends AnyShape, Pointer> extends Shape<Input<ProvidedShape>, Output<ProvidedShape> | Pointer> implements DeepPartialProtocol<LazyShape<DeepPartialShape<ProvidedShape>, Pointer>> {
    /**
     * The provider callback that returns the shape to which {@link LazyShape} delegates input handling.
     */
    readonly shapeProvider: () => ProvidedShape;
    /**
     * The provider callback that returns the value that is used instead of a circular reference.
     */
    readonly pointerProvider: (value: Input<ProvidedShape>, options: ApplyOptions) => Pointer;
    /**
     * The provider that caches the returned shape.
     */
    private _cachingShapeProvider;
    /**
     * The map from nonce to an array of inputs seen during parsing.
     */
    private _stackMap;
    /**
     * Creates a new {@link LazyShape} instance.
     *
     * @param shapeProvider The provider callback that returns the shape to which {@link LazyShape} delegates input
     * handling. The provider is called only once.
     * @param pointerProvider The provider callback that returns the value that is used instead of a circular
     * reference.
     * @template ProvidedShape The lazy-loaded shape.
     * @template Pointer The value returned when a cyclic reference is detected.
     */
    constructor(
    /**
     * The provider callback that returns the shape to which {@link LazyShape} delegates input handling.
     */
    shapeProvider: () => ProvidedShape, 
    /**
     * The provider callback that returns the value that is used instead of a circular reference.
     */
    pointerProvider: (value: Input<ProvidedShape>, options: ApplyOptions) => Pointer);
    /**
     * The lazy-loaded shape.
     */
    get providedShape(): ProvidedShape;
    at(key: unknown): AnyShape | null;
    deepPartial(): LazyShape<DeepPartialShape<ProvidedShape>, Input<DeepPartialShape<ProvidedShape>>>;
    /**
     * Replace circular references with a placeholder.
     *
     * @param pointer The value that would be used instead of circular references, or a callback that returns such a
     * value.
     * @template Pointer The value returned when a cyclic reference is detected.
     * @returns The clone of the shape.
     */
    circular<Pointer extends Any>(pointer: Pointer | ((value: Input<ProvidedShape>, options: ApplyOptions) => Pointer)): LazyShape<ProvidedShape, Pointer>;
    protected _isAsync(): boolean;
    protected _getInputs(): readonly unknown[];
    protected _clone(): this;
    protected _apply(input: unknown, options: ApplyOptions, nonce: number): Result<Output<ProvidedShape> | Pointer>;
    protected _applyAsync(input: unknown, options: ApplyOptions, nonce: number): Promise<Result<Output<ProvidedShape> | Pointer>>;
}
//# sourceMappingURL=LazyShape.d.ts.map