import { ReadonlyDict } from '../internal/objects';
import { INPUT, OUTPUT } from '../internal/shapes';
import { ApplyOptions, Issue, IssueOptions, Message, Result } from '../types';
import { EnumShape } from './EnumShape';
import { AllowShape, AnyShape, DeepPartialProtocol, DenyShape, OptionalDeepPartialShape, Shape } from './Shape';
type InferObject<PropShapes extends ReadonlyDict<AnyShape>, RestShape extends AnyShape | null, Leg extends INPUT | OUTPUT> = Prettify<UndefinedToOptional<{
    [K in keyof PropShapes]: PropShapes[K][Leg];
}> & (RestShape extends null | undefined ? {} : RestShape extends Shape ? {
    [key: string]: RestShape[Leg];
} : {})>;
type UndefinedToOptional<T> = Omit<T, OptionalKeys<T>> & {
    [K in OptionalKeys<T>]?: T[K];
};
type OptionalKeys<T> = {
    [K in keyof T]: undefined extends Extract<T[K], undefined> ? K : never;
}[keyof T];
type Prettify<T> = {
    [K in keyof T]: T[K];
} & {};
type OptionalPropShapes<PropShapes extends ReadonlyDict<AnyShape>> = {
    [K in keyof PropShapes]: AllowShape<PropShapes[K], undefined>;
};
type RequiredPropShapes<PropShapes extends ReadonlyDict<AnyShape>> = {
    [K in keyof PropShapes]: DenyShape<PropShapes[K], undefined>;
};
type DeepPartialObjectShape<PropShapes extends ReadonlyDict<AnyShape>, RestShape extends AnyShape | null> = ObjectShape<{
    [K in keyof PropShapes]: OptionalDeepPartialShape<PropShapes[K]>;
}, RestShape extends null | undefined ? null : RestShape extends Shape ? OptionalDeepPartialShape<RestShape> : RestShape>;
/**
 * Defines how unknown object keys are handled.
 *
 * - If `preserved` then unknown object keys are preserved as-is or checked with {@link ObjectShape.restShape}.
 * - If `stripped` then the input object is cloned and unknown keys are removed from it.
 * - If `exact` then an issue is raised if an unknown key is met.
 *
 * @see {@link ObjectShape.keysMode}
 * @group Other
 */
export type ObjectKeysMode = 'preserved' | 'stripped' | 'exact';
/**
 * The shape of an object.
 *
 * @template PropShapes The mapping from a string object key to a corresponding value shape.
 * @template RestShape The shape that constrains values of
 * [a string index signature](https://www.typescriptlang.org/docs/handbook/2/objects.html#index-signatures), or `null`
 * if there's no index signature.
 * @group Shapes
 */
export declare class ObjectShape<PropShapes extends ReadonlyDict<AnyShape>, RestShape extends AnyShape | null> extends Shape<InferObject<PropShapes, RestShape, INPUT>, InferObject<PropShapes, RestShape, OUTPUT>> implements DeepPartialProtocol<DeepPartialObjectShape<PropShapes, RestShape>> {
    /**
     * The mapping from an object key to a corresponding value shape.
     */
    readonly propShapes: PropShapes;
    /**
     * The shape that constrains values of
     * [a string index signature](https://www.typescriptlang.org/docs/handbook/2/objects.html#index-signatures).
     *
     * @see {@link ObjectShape.rest}
     */
    readonly restShape: RestShape;
    /**
     * The mode of unknown keys handling.
     */
    readonly keysMode: ObjectKeysMode;
    /**
     * The array of known object keys.
     */
    readonly keys: readonly string[];
    /**
     * The array of property shapes, parallel to {@link ObjectShape.keys}.
     */
    readonly valueShapes: readonly Shape[];
    /**
     * The issue options or the issue message.
     */
    protected _options: IssueOptions | Message | undefined;
    /**
     * Returns issues associated with an invalid input value type.
     */
    protected _typeIssueFactory: (input: unknown, options: ApplyOptions) => Issue;
    /**
     * Returns issues which describe that an object has unknown properties.
     */
    protected _exactIssueFactory?: (input: unknown, options: ApplyOptions, param: unknown) => Issue;
    /**
     * Creates a new {@link ObjectShape} instance.
     *
     * @param propShapes The mapping from an object key to a corresponding value shape.
     * @param restShape The shape that constrains values of
     * [a string index signature](https://www.typescriptlang.org/docs/handbook/2/objects.html#index-signatures). If `null`
     * then values thea fall under the index signature are unconstrained.
     * @param options The issue options or the issue message.
     * @param keysMode The mode of keys handling.
     * @template PropShapes The mapping from an object key to a corresponding value shape.
     * @template RestShape The shape that constrains values of
     * [a string index signature](https://www.typescriptlang.org/docs/handbook/2/objects.html#index-signatures).
     */
    constructor(
    /**
     * The mapping from an object key to a corresponding value shape.
     */
    propShapes: PropShapes, 
    /**
     * The shape that constrains values of
     * [a string index signature](https://www.typescriptlang.org/docs/handbook/2/objects.html#index-signatures).
     *
     * @see {@link ObjectShape.rest}
     */
    restShape: RestShape, options?: IssueOptions | Message, 
    /**
     * The mode of unknown keys handling.
     */
    keysMode?: ObjectKeysMode);
    /**
     * The enum shape that describes object keys.
     */
    get keysShape(): EnumShape<keyof PropShapes>;
    at(key: any): AnyShape | null;
    /**
     * Merge properties from the other object shape.
     *
     * **Note:** This method returns a shape without any operations.
     *
     * If a property with the same key already exists on this object shape then it is overwritten. The index signature of
     * this shape and its {@link ObjectShape.keysMode} is preserved intact.
     *
     * @param shape The object shape which properties must be added to this object shape.
     * @returns The new object shape.
     * @template T Properties to add.
     */
    extend<T extends ReadonlyDict<AnyShape>>(shape: ObjectShape<T, any>): ObjectShape<Omit<PropShapes, keyof T> & T, RestShape>;
    /**
     * Add properties to an object shape.
     *
     * **Note:** This method returns a shape without any operations.
     *
     * If a property with the same key already exists on this object shape then it is overwritten. The index signature of
     * this shape and its {@link ObjectShape.keysMode} is preserved intact.
     *
     * @param shapes The properties to add.
     * @returns The new object shape.
     * @template T The shapes of properties to add.
     */
    extend<T extends ReadonlyDict<AnyShape>>(shapes: T): ObjectShape<Omit<PropShapes, keyof T> & T, RestShape>;
    /**
     * Returns an object shape that only has properties with listed keys.
     *
     * **Note:** This method returns a shape without any operations.
     *
     * @param keys The array of property keys to pick.
     * @returns The new object shape.
     * @template K The tuple of keys to pick.
     */
    pick<K extends ReadonlyArray<keyof PropShapes>>(keys: K): ObjectShape<Pick<PropShapes, K[number]>, RestShape>;
    /**
     * Returns an object shape that doesn't have the listed keys.
     *
     * **Note:** This method returns a shape without any operations.
     *
     * @param keys The array of property keys to omit.
     * @returns The new object shape.
     * @template K The tuple of keys to omit.
     */
    omit<K extends ReadonlyArray<keyof PropShapes>>(keys: K): ObjectShape<Omit<PropShapes, K[number]>, RestShape>;
    /**
     * Returns an object shape with all properties marked as optional.
     *
     * **Note:** This method returns a shape without any operations.
     *
     * @returns The new object shape.
     */
    partial(): ObjectShape<OptionalPropShapes<PropShapes>, RestShape>;
    /**
     * Returns an object shape with keys marked as optional.
     *
     * **Note:** This method returns a shape without any operations.
     *
     * @param keys The array of property keys to make optional.
     * @returns The new object shape.
     * @template K The array of string keys.
     */
    partial<K extends ReadonlyArray<keyof PropShapes>>(keys: K): ObjectShape<Omit<PropShapes, K[number]> & OptionalPropShapes<Pick<PropShapes, K[number]>>, RestShape>;
    deepPartial(): DeepPartialObjectShape<PropShapes, RestShape>;
    /**
     * Returns an object shape with all properties marked as required.
     *
     * @returns The new object shape.
     */
    required(): ObjectShape<RequiredPropShapes<PropShapes>, RestShape>;
    /**
     * Returns an object shape with keys marked as required.
     *
     * @param keys The array of property keys to make required.
     * @returns The new object shape.
     * @template K The array of string keys.
     */
    required<K extends ReadonlyArray<keyof PropShapes>>(keys: K): ObjectShape<Omit<PropShapes, K[number]> & RequiredPropShapes<Pick<PropShapes, K[number]>>, RestShape>;
    /**
     * Returns an object shape that allows only known keys and has no index signature.
     *
     * @param options The issue options or the issue message.
     * @returns The new object shape.
     */
    exact(options?: IssueOptions | Message): ObjectShape<PropShapes, null>;
    /**
     * Returns an object shape that doesn't have an index signature and all unknown keys are stripped.
     *
     * @returns The new object shape.
     */
    strip(): ObjectShape<PropShapes, null>;
    /**
     * Returns an object shape that has an index signature that doesn't constrain values.
     *
     * @returns The new object shape.
     */
    preserve(): ObjectShape<PropShapes, null>;
    /**
     * Returns an object shape that has an index signature that is constrained by the given shape.
     *
     * @param restShape The shape that constrains values of
     * [a string index signature](https://www.typescriptlang.org/docs/handbook/2/objects.html#index-signatures), or `null`
     * if there's no index signature.
     * @returns The new object shape.
     * @template S The index signature shape.
     */
    rest<S extends AnyShape | null>(restShape: S): ObjectShape<PropShapes, S>;
    protected _isAsync(): boolean;
    protected _getInputs(): readonly unknown[];
    protected _apply(input: any, options: ApplyOptions, nonce: number): Result<InferObject<PropShapes, RestShape, OUTPUT>>;
    protected _applyAsync(input: any, options: ApplyOptions, nonce: number): Promise<Result<InferObject<PropShapes, RestShape, OUTPUT>>>;
    /**
     * Unknown keys are preserved as is and aren't checked.
     */
    private _applyRestUnchecked;
    /**
     * Unknown keys are either parsed with a {@link ObjectShape.restShape}, stripped, or cause an issue.
     */
    private _applyRestChecked;
}
export {};
//# sourceMappingURL=ObjectShape.d.ts.map