import { Dict, ReadonlyDict } from '../internal/objects';
import type { INPUT, OUTPUT } from '../internal/shapes';
import { Promisify } from '../internal/shapes';
import { Type } from '../Type';
import { Any, ApplyOperationsCallback, ApplyOptions, CheckResult, Err, Issue, IssueOptions, Message, Ok, Operation, OperationCallback, OperationOptions, ParameterizedOperationOptions, ParameterizedRefineOptions, ParseOptions, RefineOptions, RefinePredicate, Result } from '../types';
export declare const unknownInputs: readonly Type<"unknown">[];
/**
 * Extracts the shape input type.
 *
 * @template S The shape from which the input type must be inferred.
 * @group Type Inference
 */
export type Input<S extends AnyShape> = S[INPUT];
/**
 * Extracts the shape output type.
 *
 * @template S The shape from which the output type must be inferred.
 * @group Type Inference
 */
export type Output<S extends AnyShape> = S[OUTPUT];
/**
 * Excludes `U` from `T` only if `U` is a literal type.
 */
type ExcludeLiteral<T, U> = number extends U ? T : string extends U ? T : symbol extends U ? T : bigint extends U ? T : object extends U ? T : boolean extends U ? T : Exclude<T, U>;
/**
 * The ephemeral unique symbol that is used for type branding by {@link Branded}.
 */
declare const BRAND: unique symbol;
/**
 * The branded type.
 *
 * @template Value The value to brand.
 * @template Brand The brand value.
 * @group Other
 */
export type Branded<Value, Brand> = Value & {
    [BRAND]: Brand;
};
/**
 * A shape should implement {@link DeepPartialProtocol} to support conversion to a deep partial alternative.
 *
 * @template S The deep partial alternative of the shape.
 * @group Other
 */
export interface DeepPartialProtocol<S extends AnyShape> {
    /**
     * Converts the shape and its child shapes to deep partial alternatives.
     *
     * **Note:** This method returns a shape without any operations.
     *
     * @returns The deep partial clone of the shape.
     */
    deepPartial(): S;
}
/**
 * An arbitrary shape.
 *
 * @group Shapes
 */
export type AnyShape = Shape | Shape<never>;
/**
 * Returns the shape that refines the shape output.
 *
 * @template S The base shape.
 * @template RefinedValue The refined output value.
 * @group Shapes
 */
export type RefineShape<S extends AnyShape, RefinedValue> = Shape<Input<S>, RefinedValue> & Omit<S, keyof Shape>;
/**
 * Returns the deep partial alternative of the shape if it implements {@link DeepPartialProtocol}, or returns shape as
 * is if it doesn't.
 *
 * @template S The shape to convert to a deep partial alternative.
 * @group Shapes
 */
export type DeepPartialShape<S extends AnyShape> = S extends DeepPartialProtocol<infer T> ? T : S;
/**
 * Shape that is both optional and deep partial.
 *
 * @template S The shape to convert to an optional deep partial alternative.
 * @group Shapes
 */
export type OptionalDeepPartialShape<S extends AnyShape> = AllowShape<DeepPartialShape<S>, undefined>;
/**
 * Shortcut for {@link ReplaceShape} that allows the same value as both an input and an output.
 *
 * @template S The shape that parses the input without the replaced value.
 * @template AllowedValue The value that is allowed as an input and output.
 * @group Shapes
 */
export type AllowShape<S extends AnyShape, AllowedValue> = ReplaceShape<S, AllowedValue, AllowedValue>;
/**
 * Shortcut for {@link ExcludeShape} that doesn't impose the exclusion on the type level.
 *
 * @template BaseShape The base shape.
 * @template ExcludedShape The shape to which the output must not conform.
 * @group Shapes
 */
export interface NotShape<BaseShape extends AnyShape, ExcludedShape extends AnyShape> extends Shape<Input<BaseShape>, Output<BaseShape>>, DeepPartialProtocol<NotShape<DeepPartialShape<BaseShape>, ExcludedShape>> {
    /**
     * The base shape.
     */
    readonly baseShape: BaseShape;
    /**
     * The shape to which the output must not conform.
     */
    readonly excludedShape: ExcludedShape;
}
/**
 * The baseline shape implementation.
 *
 * @template InputValue The input value.
 * @template OutputValue The output value.
 * @group Shapes
 */
export declare class Shape<InputValue = any, OutputValue = InputValue> {
    /**
     * The mapping from an issue type to a corresponding issue message.
     */
    static readonly messages: import("../messages").Messages;
    /**
     * The dictionary of shape annotations.
     *
     * @see {@link Shape.annotate}
     */
    annotations: Dict;
    /**
     * The array of operations that are applied to the shape output.
     *
     * @see {@link Shape.addOperation}
     * @see [Operations](https://github.com/smikhalevski/doubter#operations)
     */
    operations: readonly Operation[];
    /**
     * The callback that applies {@link Shape.operations} to the shape output value.
     *
     * This method returns a promise if there are async {@link Shape.operations}.
     *
     * If the shape overrides only {@link Shape._apply} and doesn't override {@link Shape._applyAsync} then it's only safe
     * to call this method _as the last statement_ in {@link Shape._apply}. Otherwise, it may return an unexpected
     * promise.
     *
     * If the shape overrides both {@link Shape._apply} and {@link Shape._applyAsync} then this method would always
     * synchronously return a {@link Result} inside {@link Shape._apply}.
     */
    protected _applyOperations: ApplyOperationsCallback;
    /**
     * Returns a sub-shape that describes a value associated with the given property name, or `null` if there's no such
     * sub-shape.
     *
     * @param key The key for which the sub-shape must be retrieved.
     * @returns The sub-shape or `null` if there's no such key in the shape.
     */
    at(key: unknown): AnyShape | null;
    /**
     * Returns `true` if the shape accepts given input type or value, or `false` otherwise.
     *
     * @param input The type or value that must be checked.
     */
    accepts(input: unknown): boolean;
    /**
     * Adds annotations to the shape.
     *
     * @param annotations Annotations to add.
     * @returns The clone of the shape with the updated annotations.
     * @see {@link Shape.annotations}
     */
    annotate(annotations: ReadonlyDict): this;
    /**
     * Adds a synchronous operation to the shape.
     *
     * @param cb The callback that synchronously applies an operation to the shape output value.
     * @param options The operation options.
     * @returns The clone of the shape.
     * @template Param The param that is passed to the operation when it is applied.
     * @see [Operations](https://github.com/smikhalevski/doubter#operations)
     */
    addOperation<Param>(cb: OperationCallback<Result<OutputValue>, OutputValue, Param>, options?: ParameterizedOperationOptions<Param>): this;
    /**
     * Adds a synchronous operation to the shape.
     *
     * @param cb The callback that synchronously applies an operation to the shape output value.
     * @param options The operation options.
     * @returns The clone of the shape.
     * @see [Operations](https://github.com/smikhalevski/doubter#operations)
     */
    addOperation(cb: OperationCallback<Result<OutputValue>, OutputValue>, options?: OperationOptions): this;
    /**
     * Adds an asynchronous operation to the shape.
     *
     * @param cb The callback that asynchronously applies an operation to the shape output value.
     * @param options The operation options.
     * @returns The clone of the shape.
     * @template Param The param that is passed to the operation when it is applied.
     * @see [Operations](https://github.com/smikhalevski/doubter#operations)
     */
    addAsyncOperation<Param>(cb: OperationCallback<PromiseLike<Result<OutputValue>>, OutputValue, Param>, options?: ParameterizedOperationOptions<Param>): this;
    /**
     * Adds an asynchronous operation to the shape.
     *
     * @param cb The callback that asynchronously applies an operation to the shape output value.
     * @param options The operation options.
     * @returns The clone of the shape.
     * @see [Operations](https://github.com/smikhalevski/doubter#operations)
     */
    addAsyncOperation(cb: OperationCallback<PromiseLike<Result<OutputValue>>, OutputValue>, options?: OperationOptions): this;
    /**
     * Adds {@link Shape.addOperation a synchronous operation} that checks that the shape output satisfies a requirement.
     *
     * The callback must return `null` or `undefined` if value is valid, or an {@link Issue issue} or an array of issues
     * if value is invalid. If a callback returns an empty array, it is considered that no issues have occurred.
     *
     * @param cb The callback that checks that a value satisfies a requirement and returns issues if it doesn't.
     * @param options The operation options.
     * @returns The clone of the shape.
     * @template Param The param that is passed to the operation when it is applied.
     * @see {@link Shape.refine}
     * @see {@link Shape.checkAsync}
     * @see [Checks](https://github.com/smikhalevski/doubter#checks)
     */
    check<Param>(cb: OperationCallback<CheckResult, OutputValue, Param>, options: ParameterizedOperationOptions<Param>): this;
    /**
     * Adds {@link Shape.addOperation a synchronous operation} that checks that the shape output satisfies a requirement.
     *
     * The callback must return `null` or `undefined` if value is valid, or an {@link Issue issue} or an array of issues
     * if value is invalid. If a callback returns an empty array, it is considered that no issues have occurred.
     *
     * @param cb The callback that checks that a value satisfies a requirement and returns issues if it doesn't.
     * @param options The operation options.
     * @returns The clone of the shape.
     * @see {@link Shape.refine}
     * @see {@link Shape.checkAsync}
     * @see [Checks](https://github.com/smikhalevski/doubter#checks)
     */
    check(cb: OperationCallback<CheckResult, OutputValue>, options?: OperationOptions): this;
    /**
     * Adds {@link Shape.addAsyncOperation an asynchronous operation} that checks that the shape output satisfies a
     * requirement.
     *
     * The callback must return a promise that is resolved with `null` or `undefined` if value is valid, or an
     * {@link Issue issue} or an array of issues if value is invalid. If promise resolves with an empty array, it is
     * considered that no issues have occurred.
     *
     * @param cb The callback that checks that a value satisfies a requirement and returns issues if it doesn't.
     * @param options The operation options.
     * @returns The clone of the shape.
     * @template Param The param that is passed to the operation when it is applied.
     * @see {@link Shape.check}
     * @see {@link Shape.refineAsync}
     * @see [Checks](https://github.com/smikhalevski/doubter#checks)
     */
    checkAsync<Param>(cb: OperationCallback<PromiseLike<CheckResult>, OutputValue, Param>, options: ParameterizedOperationOptions<Param>): this;
    /**
     * Adds {@link Shape.addAsyncOperation an asynchronous operation} that checks that the shape output satisfies a
     * requirement.
     *
     * The callback must return a promise that is resolved with `null` or `undefined` if value is valid, or an
     * {@link Issue issue} or an array of issues if value is invalid. If promise resolves with an empty array, it is
     * considered that no issues have occurred.
     *
     * @param cb The callback that checks that a value satisfies a requirement and returns issues if it doesn't.
     * @param options The operation options.
     * @returns The clone of the shape.
     * @see {@link Shape.check}
     * @see {@link Shape.refineAsync}
     * @see [Checks](https://github.com/smikhalevski/doubter#checks)
     */
    checkAsync(cb: OperationCallback<PromiseLike<CheckResult>, OutputValue>, options?: OperationOptions): this;
    /**
     * Adds {@link Shape.addOperation a synchronous operation} that refines the shape output type with the
     * [narrowing predicate](https://www.typescriptlang.org/docs/handbook/2/narrowing.html).
     *
     * @param cb The predicate that returns `true` if the value conforms the required type, or `false` otherwise.
     * @param options The operation options or the issue message.
     * @returns The shape with the narrowed output.
     * @template RefinedValue The narrowed output value.
     * @template Param The param that is passed to the operation when it is applied.
     * @see {@link Shape.check}
     * @see {@link Shape.refineAsync}
     * @see [Refinements](https://github.com/smikhalevski/doubter#refinements)
     */
    refine<RefinedValue extends OutputValue, Param>(cb: RefinePredicate<OutputValue, RefinedValue, Param>, options: ParameterizedRefineOptions<Param> | Message): RefineShape<this, RefinedValue>;
    /**
     * Adds {@link Shape.addOperation a synchronous operation} that refines the shape output type with the
     * [narrowing predicate](https://www.typescriptlang.org/docs/handbook/2/narrowing.html).
     *
     * @param cb The predicate that returns `true` if the value conforms the required type, or `false` otherwise.
     * @param options The operation options or the issue message.
     * @returns The shape with the narrowed output.
     * @template RefinedValue The narrowed output value.
     * @see {@link Shape.check}
     * @see {@link Shape.refineAsync}
     * @see [Refinements](https://github.com/smikhalevski/doubter#refinements)
     */
    refine<RefinedValue extends OutputValue>(cb: RefinePredicate<OutputValue, RefinedValue>, options?: RefineOptions | Message): RefineShape<this, RefinedValue>;
    /**
     * Adds {@link Shape.addOperation a synchronous operation} that checks that the output value conforms the predicate.
     *
     * @param cb The predicate that returns truthy result if the value is valid, or returns falsy result otherwise.
     * @param options The operation options or the issue message.
     * @returns The clone of the shape.
     * @template Param The param that is passed to the operation when it is applied.
     * @see {@link Shape.check}
     * @see {@link Shape.refineAsync}
     * @see [Refinements](https://github.com/smikhalevski/doubter#refinements)
     */
    refine<Param>(cb: OperationCallback<unknown, OutputValue, Param>, options?: ParameterizedRefineOptions<Param> | Message): this;
    /**
     * Adds {@link Shape.addOperation a synchronous operation} that checks that the output value conforms the predicate.
     *
     * @param cb The predicate that returns truthy result if the value is valid, or returns falsy result otherwise.
     * @param options The operation options or the issue message.
     * @returns The clone of the shape.
     * @see {@link Shape.check}
     * @see {@link Shape.refineAsync}
     * @see [Refinements](https://github.com/smikhalevski/doubter#refinements)
     */
    refine(cb: OperationCallback<unknown, OutputValue>, options?: RefineOptions | Message): this;
    /**
     * Adds {@link Shape.addAsyncOperation an asynchronous operation} that checks that the output value conforms the
     * predicate.
     *
     * @param cb The predicate that returns a promise that resolves with a truthy result if the value is valid, or a falsy
     * result otherwise.
     * @param options The operation options or the issue message.
     * @returns The clone of the shape.
     * @template Param The param that is passed to the operation when it is applied.
     * @see {@link Shape.checkAsync}
     * @see {@link Shape.refine}
     * @see [Refinements](https://github.com/smikhalevski/doubter#refinements)
     */
    refineAsync<Param>(cb: OperationCallback<PromiseLike<unknown>, OutputValue, Param>, options?: ParameterizedRefineOptions<Param> | Message): this;
    /**
     * Adds {@link Shape.addAsyncOperation an asynchronous operation} that checks that the output value conforms the
     * predicate.
     *
     * @param cb The predicate that returns a promise that resolves with a truthy result if the value is valid, or a falsy
     * result otherwise.
     * @param options The operation options or the issue message.
     * @returns The clone of the shape.
     * @see {@link Shape.checkAsync}
     * @see {@link Shape.refine}
     * @see [Refinements](https://github.com/smikhalevski/doubter#refinements)
     */
    refineAsync(cb: OperationCallback<PromiseLike<unknown>, OutputValue>, options?: RefineOptions | Message): this;
    /**
     * Adds {@link Shape.addOperation a synchronous operation} that alters the output value without changing its type.
     *
     * If you want to change the base type, consider using {@link Shape.convert}.
     *
     * @param cb The callback that alters the shape output. Throw a {@link ValidationError} to notify that the alteration
     * cannot be successfully completed.
     * @param options The operation options.
     * @returns The clone of the shape.
     * @template Param The param that is passed to the operation when it is applied.
     * @see {@link Shape.convert}
     * @see {@link Shape.alterAsync}
     * @see [Alterations](https://github.com/smikhalevski/doubter#alterations)
     */
    alter<Param>(cb: OperationCallback<OutputValue, OutputValue, Param>, options: ParameterizedOperationOptions<Param>): this;
    /**
     * Adds {@link Shape.addOperation a synchronous operation} that alters the output value without changing its type.
     *
     * If you want to change the base type, consider using {@link Shape.convert}.
     *
     * @param cb The callback that alters the shape output. Throw a {@link ValidationError} to notify that the alteration
     * cannot be successfully completed.
     * @param options The operation options.
     * @returns The clone of the shape.
     * @see {@link Shape.convert}
     * @see {@link Shape.alterAsync}
     * @see [Alterations](https://github.com/smikhalevski/doubter#alterations)
     */
    alter(cb: OperationCallback<OutputValue, OutputValue>, options?: OperationOptions): this;
    /**
     * Adds {@link Shape.addAsyncOperation an asynchronous operation} that alters the output value without changing its
     * type.
     *
     * If you want to change the base type, consider using {@link Shape.convertAsync}.
     *
     * @param cb The callback that alters the shape output. Throw a {@link ValidationError} to notify that the alteration
     * cannot be successfully completed.
     * @param options The operation options.
     * @returns The clone of the shape.
     * @template Param The param that is passed to the operation when it is applied.
     * @see {@link Shape.convertAsync}
     * @see {@link Shape.alter}
     * @see [Alterations](https://github.com/smikhalevski/doubter#alterations)
     */
    alterAsync<Param>(cb: OperationCallback<PromiseLike<OutputValue>, OutputValue, Param>, options: ParameterizedOperationOptions<Param>): this;
    /**
     * Adds {@link Shape.addAsyncOperation an asynchronous operation} that alters the output value without changing its
     * type.
     *
     * If you want to change the base type, consider using {@link Shape.convertAsync}.
     *
     * @param cb The callback that alters the shape output. Throw a {@link ValidationError} to notify that the alteration
     * cannot be successfully completed.
     * @param options The operation options.
     * @returns The clone of the shape.
     * @see {@link Shape.convertAsync}
     * @see {@link Shape.alter}
     * @see [Alterations](https://github.com/smikhalevski/doubter#alterations)
     */
    alterAsync(cb: OperationCallback<PromiseLike<OutputValue>, OutputValue>, options?: OperationOptions): this;
    /**
     * Synchronously converts the output value of the shape.
     *
     * If you want to don't want to change the base type, consider using {@link Shape.alter}.
     *
     * @param cb The callback that converts the input value. Throw a {@link ValidationError} to notify that the conversion
     * cannot be successfully completed.
     * @returns The {@link ConvertShape} instance.
     * @template ConvertedValue The value returned from the callback that converts the output value of this shape.
     * @see {@link Shape.convertAsync}
     * @see {@link Shape.alter}
     */
    convert<ConvertedValue>(
    /**
     * @param value The shape output value.
     * @param options Parsing options.
     */
    cb: (value: OutputValue, options: ApplyOptions) => ConvertedValue): Shape<InputValue, ConvertedValue>;
    /**
     * Asynchronously converts the output value of the shape.
     *
     * If you want to don't want to change the base type, consider using {@link Shape.alterAsync}.
     *
     * @param cb The callback that converts the input value asynchronously. The returned promise can be rejected with a
     * {@link ValidationError} to notify that the conversion cannot be successfully completed.
     * @returns The {@link ConvertShape} instance.
     * @template ConvertedValue The value returned from the callback that converts the output value of this shape.
     * @see {@link Shape.convert}
     * @see {@link Shape.alterAsync}
     */
    convertAsync<ConvertedValue>(
    /**
     * @param value The shape output value.
     * @param options Parsing options.
     */
    cb: (value: OutputValue, options: ApplyOptions) => PromiseLike<ConvertedValue>): Shape<InputValue, ConvertedValue>;
    /**
     * Pipes the output of this shape to the input of another shape.
     *
     * @param shape The shape that validates the output if this shape.
     * @template OutputShape The output value.
     */
    to<OutputShape extends AnyShape>(shape: OutputShape): PipeShape<this, OutputShape>;
    /**
     * Returns a shape that adds a brand to the output type.
     *
     * @returns A shape with the branded output type.
     * @template Brand The brand value.
     */
    brand<Brand extends string | number | symbol>(): RefineShape<this, Branded<OutputValue, Brand>>;
    /**
     * Replaces an input value with an output value.
     *
     * @param inputValue The input value to replace.
     * @param outputValue The output value that is returned if an `inputValue` is received.
     * @template InputValue The input value to replace.
     * @template OutputValue The output value that is used as the replacement for an input value.
     */
    replace<InputValue extends Any, OutputValue extends Any>(inputValue: InputValue, outputValue: OutputValue): ReplaceShape<this, InputValue, OutputValue>;
    /**
     * Allows an input value, so it is passed directly to the output.
     *
     * @param value The allowed value.
     * @template AllowedValue The allowed value.
     */
    allow<AllowedValue extends Any>(value: AllowedValue): AllowShape<this, AllowedValue>;
    /**
     * Excludes value from both input and output.
     *
     * @param value The excluded value.
     * @param options The issue options or the issue message.
     * @template DeniedValue The denied value.
     */
    deny<DeniedValue extends InputValue | OutputValue>(value: DeniedValue, options?: IssueOptions | Message): DenyShape<this, DeniedValue>;
    /**
     * Replaces `undefined` input value with an `undefined` output value.
     */
    optional(): AllowShape<this, undefined>;
    /**
     * Replaces `undefined` input value with a default output value.
     *
     * @param defaultValue The value that should be used if an input value is `undefined`.
     * @template DefaultValue The value that is used as the replacement for `undefined`.
     */
    optional<DefaultValue extends Any>(defaultValue: DefaultValue): ReplaceShape<this, undefined, DefaultValue>;
    /**
     * Replaces `null` input value with an `null` output value.
     */
    nullable(): AllowShape<this, null>;
    /**
     * Replaces `null` input value with a default output value.
     *
     * @param defaultValue The value that should be used if an input value is `null`.
     * @template DefaultValue The value that is used as the replacement for `null`.
     */
    nullable<DefaultValue extends Any>(defaultValue: DefaultValue): ReplaceShape<this, null, DefaultValue>;
    /**
     * Passes `null` and `undefined` input values directly to the output without parsing.
     */
    nullish(): AllowShape<AllowShape<this, null>, undefined>;
    /**
     * Replaces `null` and `undefined` input value with a default output value.
     *
     * @param defaultValue The value that should be used if an input value is `undefined` or `null`.
     * @template DefaultValue The value that is used as the replacement for `undefined` and `null`.
     */
    nullish<DefaultValue extends Any>(defaultValue?: DefaultValue): ReplaceShape<ReplaceShape<this, null, DefaultValue>, undefined, DefaultValue>;
    /**
     * Prevents an input and output from being `undefined`.
     *
     * @param options The issue options or the issue message.
     */
    nonOptional(options?: IssueOptions | Message): DenyShape<this, undefined>;
    /**
     * Returns `undefined` if parsing fails.
     */
    catch(): CatchShape<this, undefined>;
    /**
     * Returns the fallback value if parsing fails.
     *
     * @param fallback The value or a callback that returns a value that is returned if parsing has failed. A callback
     * receives an input value, an array of raised issues, and {@link ApplyOptions parsing options}.
     * @template FallbackValue The fallback value.
     */
    catch<FallbackValue extends Any>(fallback: FallbackValue | ((input: any, issues: Issue[], options: ApplyOptions) => FallbackValue)): CatchShape<this, FallbackValue>;
    /**
     * Checks that the input doesn't match the shape.
     *
     * @param shape The shape to which the output must not conform.
     * @param options The issue options or the issue message.
     * @template ExcludedShape The shape to which the output must not conform.
     */
    exclude<ExcludedShape extends AnyShape>(shape: ExcludedShape, options?: IssueOptions | Message): ExcludeShape<this, ExcludedShape>;
    /**
     * Checks that the input doesn't match the shape.
     *
     * This method works exactly as {@link Shape.exclude} at runtime, but it doesn't perform the exclusion on the type
     * level.
     *
     * @param shape The shape to which the output must not conform.
     * @param options The issue options or the issue message.
     * @template ExcludedShape The shape to which the output must not conform.
     */
    not<ExcludedShape extends AnyShape>(shape: ExcludedShape, options?: IssueOptions | Message): NotShape<this, ExcludedShape>;
    /**
     * Must return `true` if the shape must be used in async context only, otherwise the shape can be used in both sync
     * and async contexts. Override this method to implement a custom shape.
     *
     * @see [Advanced shapes](https://github.com/smikhalevski/doubter#advanced-shapes)
     */
    protected _isAsync(): boolean;
    /**
     * Returns input types and literal values that this shape can accept as an input.
     *
     * @see [Advanced shapes](https://github.com/smikhalevski/doubter#advanced-shapes)
     */
    protected _getInputs(): readonly unknown[];
    /**
     * Clones the shape.
     *
     * @see [Advanced shapes](https://github.com/smikhalevski/doubter#advanced-shapes)
     */
    protected _clone(): this;
    /**
     * Synchronously parses the input.
     *
     * @param input The shape input to parse.
     * @param options Parsing options.
     * @param nonce The globally unique number that identifies the parsing process.
     * @returns `null` if input matches the output, {@link Ok} that wraps the output, or an array of captured issues.
     * @see [Advanced shapes](https://github.com/smikhalevski/doubter#advanced-shapes)
     */
    protected _apply(input: unknown, options: ApplyOptions, nonce: number): Result<OutputValue>;
    /**
     * Asynchronously parses the input.
     *
     * @param input The shape input to parse.
     * @param options Parsing options.
     * @param nonce The globally unique number that identifies the parsing process.
     * @returns `null` if input matches the output, {@link Ok} that wraps the output, or an array of captured issues.
     * @see [Advanced shapes](https://github.com/smikhalevski/doubter#advanced-shapes)
     */
    protected _applyAsync(input: unknown, options: ApplyOptions, nonce: number): Promise<Result<OutputValue>>;
}
export interface Shape<InputValue, OutputValue> {
    /**
     * The shape input type.
     *
     * @internal
     */
    readonly [INPUT]: InputValue;
    /**
     * The shape output type.
     *
     * @internal
     */
    readonly [OUTPUT]: OutputValue;
    /**
     * The array of unique input types and values that are accepted by the shape.
     */
    readonly inputs: readonly unknown[];
    /**
     * `true` if the shape allows only {@link Shape.parseAsync} and throws an error if {@link Shape.parse} is called, or
     * `false` if the shape can be used in both sync and async contexts.
     */
    readonly isAsync: boolean;
    /**
     * Synchronously parses the value and returns {@link Ok} or {@link Err} object that wraps the result.
     *
     * @param input The value to parse.
     * @param options Parsing options.
     * @returns The {@link Ok} instance if parsing has succeeded or {@link Err} if parsing has failed.
     * @throws {@link !Error Error} if the shape doesn't support the sync parsing, see {@link Shape.isAsync}.
     */
    try(input: unknown, options?: ApplyOptions): Ok<OutputValue> | Err;
    /**
     * Asynchronously parses the value and returns {@link Ok} or {@link Err} object that wraps the result.
     *
     * @param input The value to parse.
     * @param options Parsing options.
     * @returns The {@link Ok} instance if parsing has succeeded or {@link Err} if parsing has failed.
     */
    tryAsync(input: unknown, options?: ApplyOptions): Promise<Ok<OutputValue> | Err>;
    /**
     * Synchronously parses the value.
     *
     * @param input The value to parse.
     * @param options Parsing options.
     * @returns The value that conforms the output type of the shape.
     * @throws {@link !Error Error} if the shape doesn't support the sync parsing, see {@link Shape.isAsync}.
     * @throws {@link ValidationError} if any issues occur during parsing.
     */
    parse(input: unknown, options?: ParseOptions): OutputValue;
    /**
     * Asynchronously parses the value.
     *
     * @param input The value to parse.
     * @param options Parsing options.
     * @returns The promise that resolves with the value that conforms the output type of the shape, or rejects with a
     * {@link ValidationError} if any issues occur during parsing.
     */
    parseAsync(input: unknown, options?: ParseOptions): Promisify<OutputValue>;
    /**
     * Synchronously parses the value and returns `undefined` if parsing fails.
     *
     * @param input The value to parse.
     * @returns The value that conforms the output type of the shape.
     * @throws {@link !Error Error} if the shape doesn't support the sync parsing, see {@link Shape.isAsync}.
     */
    parseOrDefault(input: unknown): OutputValue | undefined;
    /**
     * Synchronously parses the value and returns the default value if parsing fails.
     *
     * @param input The value to parse.
     * @param defaultValue The default value that is returned if parsing fails.
     * @param options Parsing options.
     * @template DefaultValue The default value that is returned if parsing fails.
     * @returns The value that conforms the output type of the shape.
     * @throws {@link !Error Error} if the shape doesn't support the sync parsing, see {@link Shape.isAsync}.
     */
    parseOrDefault<DefaultValue>(input: unknown, defaultValue: DefaultValue, options?: ApplyOptions): OutputValue | DefaultValue;
    /**
     * Asynchronously parses the value and returns `undefined` value if parsing fails.
     *
     * @param input The value to parse.
     * @returns The value that conforms the output type of the shape.
     */
    parseOrDefaultAsync(input: unknown): Promisify<OutputValue | undefined>;
    /**
     * Asynchronously parses the value and returns the default value if parsing fails.
     *
     * @param input The value to parse.
     * @param defaultValue The default value that is returned if parsing fails.
     * @param options Parsing options.
     * @template DefaultValue The default value that is returned if parsing fails.
     * @returns The value that conforms the output type of the shape.
     */
    parseOrDefaultAsync<DefaultValue>(input: unknown, defaultValue: DefaultValue, options?: ApplyOptions): Promisify<OutputValue | DefaultValue>;
}
/**
 * The shape that applies a converter to the input.
 *
 * @template ConvertedValue The output value of the callback that converts the input value.
 * @group Shapes
 */
export declare class ConvertShape<ConvertedValue> extends Shape<any, ConvertedValue> {
    /**
     * The converter that is applied to the shape output value.
     *
     * @param value The input value.
     * @param options Parsing options.
     */
    readonly converter: (value: any, options: ApplyOptions) => PromiseLike<ConvertedValue> | ConvertedValue;
    /**
     * Creates the new {@link ConvertShape} instance.
     *
     * @param converter The callback that converts the input value. Converters can throw or reject with a
     * {@link ValidationError} to notify that the conversion cannot be successfully completed.
     * @param async If `true` then the convert shape waits for the promise returned from the callback to be fulfilled.
     * Otherwise, the value that is synchronously returned from the callback is used as an output.
     * @template ConvertedValue The output value of the callback that converts the input value.
     */
    constructor(
    /**
     * The converter that is applied to the shape output value.
     *
     * @param value The input value.
     * @param options Parsing options.
     */
    converter: (value: any, options: ApplyOptions) => PromiseLike<ConvertedValue> | ConvertedValue, async?: boolean);
    protected _apply(input: unknown, options: ApplyOptions, nonce: number): Result<ConvertedValue>;
    protected _applyAsync(input: unknown, options: ApplyOptions, nonce: number): Promise<Result<ConvertedValue>>;
}
/**
 * The shape that parses the output of the input shape with the output shape.
 *
 * @template InputShape The input shape.
 * @template OutputShape The output shape.
 * @group Shapes
 */
export declare class PipeShape<InputShape extends AnyShape, OutputShape extends AnyShape> extends Shape<Input<InputShape>, Output<OutputShape>> implements DeepPartialProtocol<PipeShape<DeepPartialShape<InputShape>, DeepPartialShape<OutputShape>>> {
    /**
     * The input shape.
     */
    readonly inputShape: InputShape;
    /**
     * The output shape.
     */
    readonly outputShape: OutputShape;
    /**
     * Creates the new {@link PipeShape} instance.
     *
     * @param inputShape The shape that parses the input value.
     * @param outputShape The shape that parses the output of `inputShape`.
     * @template InputShape The input shape.
     * @template OutputShape The output shape.
     */
    constructor(
    /**
     * The input shape.
     */
    inputShape: InputShape, 
    /**
     * The output shape.
     */
    outputShape: OutputShape);
    deepPartial(): PipeShape<DeepPartialShape<InputShape>, DeepPartialShape<OutputShape>>;
    protected _isAsync(): boolean;
    protected _getInputs(): readonly unknown[];
    protected _apply(input: unknown, options: ApplyOptions, nonce: number): Result<Output<OutputShape>>;
    protected _applyAsync(input: unknown, options: ApplyOptions, nonce: number): Promise<Result<Output<OutputShape>>>;
}
/**
 * The shape that replaces an input value with an output value.
 *
 * @template BaseShape The shape that parses the input without the replaced value.
 * @template InputValue The input value to replace.
 * @template OutputValue The output value that is used as the replacement for an input value.
 * @group Shapes
 */
export declare class ReplaceShape<BaseShape extends AnyShape, InputValue, OutputValue> extends Shape<Input<BaseShape> | InputValue, ExcludeLiteral<Output<BaseShape>, InputValue> | OutputValue> implements DeepPartialProtocol<ReplaceShape<DeepPartialShape<BaseShape>, InputValue, OutputValue>> {
    /**
     * The shape that parses the input without the replaced value.
     */
    readonly baseShape: BaseShape;
    /**
     * The input value to replace.
     */
    readonly inputValue: InputValue;
    /**
     * The output value that is returned if an {@link ReplaceShape.inputValue} is received.
     */
    readonly outputValue: OutputValue;
    private _result;
    /**
     * Creates the new {@link ReplaceShape} instance.
     *
     * @param baseShape The shape that parses the input without the replaced value.
     * @param inputValue The input value to replace.
     * @param outputValue The output value that is returned if an `inputValue` is received.
     * @template BaseShape The shape that parses the input without the replaced value.
     * @template InputValue The input value to replace.
     * @template OutputValue The output value that is used as the replacement for an input value.
     */
    constructor(
    /**
     * The shape that parses the input without the replaced value.
     */
    baseShape: BaseShape, 
    /**
     * The input value to replace.
     */
    inputValue: InputValue, 
    /**
     * The output value that is returned if an {@link ReplaceShape.inputValue} is received.
     */
    outputValue: OutputValue);
    deepPartial(): ReplaceShape<DeepPartialShape<BaseShape>, InputValue, OutputValue>;
    protected _isAsync(): boolean;
    protected _getInputs(): readonly unknown[];
    protected _apply(input: unknown, options: ApplyOptions, nonce: number): Result<ExcludeLiteral<Output<BaseShape>, InputValue> | OutputValue>;
    protected _applyAsync(input: unknown, options: ApplyOptions, nonce: number): Promise<Result<ExcludeLiteral<Output<BaseShape>, InputValue> | OutputValue>>;
}
/**
 * The shape that prevents both input and output from being equal to a denied value.
 *
 * @template BaseShape The shape that parses the input without the denied value.
 * @template DeniedValue The denied value.
 * @group Shapes
 */
export declare class DenyShape<BaseShape extends AnyShape, DeniedValue> extends Shape<ExcludeLiteral<Input<BaseShape>, DeniedValue>, ExcludeLiteral<Output<BaseShape>, DeniedValue>> implements DeepPartialProtocol<DenyShape<DeepPartialShape<BaseShape>, DeniedValue>> {
    /**
     * The shape that parses the input without the denied value.
     */
    readonly baseShape: BaseShape;
    /**
     * The dined value.
     */
    readonly deniedValue: DeniedValue;
    /**
     * The constraint options or an issue message.
     */
    protected _options: IssueOptions | Message | undefined;
    /**
     * Returns issues associated with an invalid input value type.
     */
    protected _typeIssueFactory: (input: unknown, options: ApplyOptions) => Issue;
    /**
     * Creates the new {@link DenyShape} instance.
     *
     * @param baseShape The shape that parses the input without the denied value.
     * @param deniedValue The dined value.
     * @param options The issue options or the issue message.
     * @template BaseShape The shape that parses the input without the denied value.
     * @template DeniedValue The denied value.
     */
    constructor(
    /**
     * The shape that parses the input without the denied value.
     */
    baseShape: BaseShape, 
    /**
     * The dined value.
     */
    deniedValue: DeniedValue, options?: IssueOptions | Message);
    deepPartial(): DenyShape<DeepPartialShape<BaseShape>, DeniedValue>;
    protected _isAsync(): boolean;
    protected _getInputs(): readonly unknown[];
    protected _apply(input: unknown, options: ApplyOptions, nonce: number): Result<ExcludeLiteral<Output<BaseShape>, DeniedValue>>;
    protected _applyAsync(input: unknown, options: ApplyOptions, nonce: number): Promise<Result<ExcludeLiteral<Output<BaseShape>, DeniedValue>>>;
}
/**
 * The shape that returns the fallback value if parsing fails.
 *
 * @template BaseShape The shape that parses the input.
 * @template FallbackValue The fallback value.
 * @group Shapes
 */
export declare class CatchShape<BaseShape extends AnyShape, FallbackValue> extends Shape<Input<BaseShape>, Output<BaseShape> | FallbackValue> implements DeepPartialProtocol<CatchShape<DeepPartialShape<BaseShape>, FallbackValue>> {
    /**
     * The shape that parses the input.
     */
    readonly baseShape: BaseShape;
    /**
     * The value or a callback that returns a value that is returned if parsing has failed.
     *
     *  A callback receives an input value, an array of raised issues, and {@link ApplyOptions parsing options}.
     */
    readonly fallback: FallbackValue | ((input: any, issues: Issue[], options: ApplyOptions) => FallbackValue);
    private _resultProvider;
    /**
     * Creates the new {@link CatchShape} instance.
     *
     * @param baseShape The shape that parses the input.
     * @param fallback The value or a callback that returns a value that is returned if parsing has failed.
     * @template BaseShape The shape that parses the input.
     * @template FallbackValue The fallback value.
     */
    constructor(
    /**
     * The shape that parses the input.
     */
    baseShape: BaseShape, 
    /**
     * The value or a callback that returns a value that is returned if parsing has failed.
     *
     *  A callback receives an input value, an array of raised issues, and {@link ApplyOptions parsing options}.
     */
    fallback: FallbackValue | ((input: any, issues: Issue[], options: ApplyOptions) => FallbackValue));
    deepPartial(): CatchShape<DeepPartialShape<BaseShape>, FallbackValue>;
    protected _isAsync(): boolean;
    protected _getInputs(): readonly unknown[];
    protected _apply(input: unknown, options: ApplyOptions, nonce: number): Result<Output<BaseShape> | FallbackValue>;
    protected _applyAsync(input: unknown, options: ApplyOptions, nonce: number): Promise<Result<Output<BaseShape> | FallbackValue>>;
}
/**
 * Checks that the input doesn't match the shape.
 *
 * @template BaseShape The base shape.
 * @template ExcludedShape The shape to which the output must not conform.
 * @group Shapes
 */
export declare class ExcludeShape<BaseShape extends AnyShape, ExcludedShape extends AnyShape> extends Shape<Input<BaseShape>, Exclude<Output<BaseShape>, Input<ExcludedShape>>> implements DeepPartialProtocol<ExcludeShape<DeepPartialShape<BaseShape>, ExcludedShape>> {
    /**
     * The base shape.
     */
    readonly baseShape: BaseShape;
    /**
     * The shape to which the output must not conform.
     */
    readonly excludedShape: ExcludedShape;
    /**
     * The constraint options or an issue message.
     */
    protected _options: IssueOptions | Message | undefined;
    /**
     * Returns issues associated with an invalid input value type.
     */
    protected _typeIssueFactory: (input: unknown, options: ApplyOptions) => Issue;
    /**
     * Creates the new {@link ExcludeShape} instance.
     *
     * @param baseShape The shape that parses the input.
     * @param excludedShape The shape to which the output must not conform.
     * @param options The issue options or the issue message.
     * @template BaseShape The base shape.
     * @template ExcludedShape The shape to which the output must not conform.
     */
    constructor(
    /**
     * The base shape.
     */
    baseShape: BaseShape, 
    /**
     * The shape to which the output must not conform.
     */
    excludedShape: ExcludedShape, options?: IssueOptions | Message);
    deepPartial(): ExcludeShape<DeepPartialShape<BaseShape>, ExcludedShape>;
    protected _isAsync(): boolean;
    protected _getInputs(): readonly unknown[];
    protected _apply(input: unknown, options: ApplyOptions, nonce: number): Result<Exclude<Output<BaseShape>, Input<ExcludedShape>>>;
    protected _applyAsync(input: unknown, options: ApplyOptions, nonce: number): Promise<Result<Exclude<Output<BaseShape>, Input<ExcludedShape>>>>;
}
export {};
//# sourceMappingURL=Shape.d.ts.map