import { ReadonlyDict } from '../internal/objects';
import { ApplyOptions, Issue, IssueOptions, Message, Result } from '../types';
import { ObjectShape } from './ObjectShape';
import { AnyShape, DeepPartialProtocol, DeepPartialShape, Input, Output, Shape } from './Shape';
type DeepPartialUnionShape<Shapes extends readonly AnyShape[]> = UnionShape<{
    [K in keyof Shapes]: DeepPartialShape<Shapes[K]>;
}>;
/**
 * The shape that requires an input to conform at least one of shapes.
 *
 * @template Shapes The array of shapes that comprise a union.
 * @group Shapes
 */
export declare class UnionShape<Shapes extends readonly AnyShape[]> extends Shape<Input<Shapes[number]>, Output<Shapes[number]>> implements DeepPartialProtocol<DeepPartialUnionShape<Shapes>> {
    /**
     * The array of shapes that comprise a union.
     */
    readonly shapes: Shapes;
    /**
     * The union constraint options or an issue message.
     */
    protected _options: IssueOptions | Message | undefined;
    /**
     * Returns issues associated with an invalid input value type.
     */
    protected _typeIssueFactory: (input: unknown, options: ApplyOptions, param: unknown) => Issue;
    /**
     * Creates a new {@link UnionShape} instance.
     *
     * @param shapes The array of shapes that comprise a union.
     * @param options The union constraint options or an issue message.
     * @template Shapes The array of shapes that comprise a union.
     */
    constructor(
    /**
     * The array of shapes that comprise a union.
     */
    shapes: Shapes, options?: IssueOptions | Message);
    at(key: unknown): AnyShape | null;
    deepPartial(): DeepPartialUnionShape<Shapes>;
    /**
     * Returns an array of shapes that should be applied to the input.
     */
    private get _lookup();
    protected _isAsync(): boolean;
    protected _getInputs(): readonly unknown[];
    protected _apply(input: unknown, options: ApplyOptions, nonce: number): Result<Output<Shapes[number]>>;
    protected _applyAsync(input: unknown, options: ApplyOptions, nonce: number): Promise<Result<Output<Shapes[number]>>>;
}
type ObjectShapeLike = ObjectShape<ReadonlyDict<AnyShape>, AnyShape | null>;
export type LookupCallback = (input: any) => readonly AnyShape[];
export interface Discriminator {
    /**
     * The discriminator property key.
     */
    key: string;
    /**
     * Values for each shape.
     */
    valueGroups: Array<ReadonlyArray<unknown>>;
}
export declare function createLookup(shapes: readonly AnyShape[]): LookupCallback;
/**
 * Creates a lookup that uses a discriminator to detect which shape to use.
 */
export declare function createDiscriminatorLookup(shapes: ObjectShapeLike[], discriminator: Discriminator): LookupCallback;
/**
 * Discriminator conforms the following rules:
 *
 * - A key that is common for all provided object shapes;
 * - Values that correspond to the key are discrete;
 * - Values uniquely identify each shape.
 */
export declare function getDiscriminator(shapes: ObjectShapeLike[]): Discriminator | null;
export {};
//# sourceMappingURL=UnionShape.d.ts.map